package Scramble::XMPP;

use strict;
use warnings;
use Carp;
use Time::HiRes;
use DateTime;
use Net::XMPP;
use YAML;
use Hash::Merge;

our $VERSION = '1.0';

sub new {
    my ($class, %args) = @_;
    my $self  = {};
    bless $self, $class;

    $self->{config} = {
        global => {
            debug => 0,
            prog  => substr($0, rindex($0, '/')+1),
           },
        server => {
            jid         => 'scramble@jabber.jp',
            password    => 'p@55w0rd',
            server_host => undef,
            server_port => 5222,
            tls         => 0,
        },
        to     => {
            emerg => [ 'hirose@jabber.jp', ],
            warn  => [ 'hirose@jabber.jp', ],
            info  => [ 'hirose@jabber.jp', ],
        },
    };

    ### override by YAML (confg or etc/scramble/xmpp.yaml)
    Hash::Merge::set_behavior( 'RIGHT_PRECEDENT' );
    my $yaml = $args{config} ? $args{config} : '/usr/klab/etc/scramble/xmpp.yaml';
    if (-r $yaml) {
        my $c = YAML::LoadFile($yaml);
        $self->{config} = Hash::Merge::merge($self->{config}, $c);
    } else {
        carp "[WARN] no such file: $yaml" unless $yaml eq 'none';
    }

    ### override by new parameter
    for my $cat (qw(global server)) {
        for (keys %{ $self->{config}{$cat} } ) {
            $self->{config}{$cat}{$_} = $args{$_} if exists $args{$_};
        }
    }
    for (qw(emerg warn info)) {
        $self->{config}{to}{$_} = [ list_uniq(@{$self->{config}{to}{$_}}) ];
    }

    ### go ahead
    $self->{connected} = 0;
    $self->{client}    = Net::XMPP::Client->new(debuglevel=>0);

    unless (index($self->{config}{server}{jid}, '@') >= 1) {
        carp '[FATAL] jid: invalid format';
        return;
    }
    my ($username, $componentname) = split /@/, $self->{config}{server}{jid};
    $self->{config}{server}{server_host} ||= $componentname;

    $self->debug(YAML::Dump($self->{config}));

    my $connectattempts = 3;
    my $connectsleep    = 1;
    my ($status, $error);
    while (--$connectattempts >= 0) {
        $status = $self->{client}->Connect(
            hostname       => $self->{config}{server}{server_host},
            port           => $self->{config}{server}{server_port},
            tls            => $self->{config}{server}{tls},
            componentname  => $componentname,
            connectiontype => 'http',
           );
        last if defined $status;
        sleep $connectsleep;
    }
    unless (defined $status) {
        carp "[FATAL] failed to connect ".join(':',$self->{config}{server}{server_host},$self->{config}{server}{server_port});
        return;
    }

    {
        # quick hack to connect Google Talk
        # override XML::Stream-1.22
        no warnings 'redefine';
        *XML::Stream::SASLClient = sub {
            my $self = shift;
            my $sid = shift;
            my $username = shift;
            my $password = shift;

            my $mechanisms = $self->GetStreamFeature($sid,"xmpp-sasl");

            return unless defined($mechanisms);

            my $sasl = new Authen::SASL(mechanism=>join(" ",@{$mechanisms}),
                                        callback=>{
                                            authname => $username."@".($self->{SIDS}->{$sid}->{to} or $self->{SIDS}->{$sid}->{hostname}),
                                            user     => $username,
                                            pass     => $password
                                           }
                                       );

            $self->{SIDS}->{$sid}->{sasl}->{client}   = $sasl->client_new();
            $self->{SIDS}->{$sid}->{sasl}->{username} = $username;
            $self->{SIDS}->{$sid}->{sasl}->{password} = $password;
            $self->{SIDS}->{$sid}->{sasl}->{authed}   = 0;
            $self->{SIDS}->{$sid}->{sasl}->{done}     = 0;

            $self->SASLAuth($sid);
        };
    }

    ($status,$error) = $self->{client}->AuthSend(
        username => $username,
        password => $self->{config}{server}{password},
        resource => 'Scramble',
       );
    unless ($status and $status eq 'ok') {
        carp "[FATAL] authentication failure";
        return;
    }

    $self->{connected} = 1;

    return $self;
}

sub send_emerg {
    my $self = shift;
    $self->_send_level('emerg', @_);
}

sub send_warn {
    my $self = shift;
    $self->_send_level('warn', @_);
}

sub send_info {
    my $self = shift;
    $self->_send_level('info', @_);
}

sub _send_level {
    my($self, $level, %param) = @_;

    $self->debug("level:$level");
    my @rcpt;
    if (ref $param{force_rcpt} eq 'ARRAY') {
        @rcpt = @{ $param{force_rcpt} };
    } elsif (ref $param{debug_rcpt} eq 'ARRAY') {
        @rcpt = @{ $param{debug_rcpt} };
    } else {
        @rcpt = @{ $self->{config}{to}{$level} };
        push @rcpt, @{ $param{'rcpt'} } if ref $param{'rcpt'} eq 'ARRAY';
    }

    $param{'to'}      = [ @rcpt ];
    $param{'subject'} = uc($level).($param{'subject'} ? ' '.$param{'subject'} : '');
    $self->_send_message(%param);
}

# %param
#   to   => []
#   body => '' # must be decoded Perl's internal string.
#  (subject => '') for compatibility ScrambleMail. append to body
sub _send_message {
    my $self = shift;
    my %param = @_;
    my $mesg;
    my @recip = @{
        ref $param{'to'} eq 'ARRAY'
            ? $param{'to'}
            : ['hirose@jabber.jp']
        };
    $self->debug('recip:'.join(' ',@recip));

    my $dt = DateTime->from_epoch( epoch => Time::HiRes::time,
                                   time_zone => 'Asia/Tokyo',
                                 );
    my $timestamp = $dt->strftime("%F %T.%9N%z");
    my $caller;
    for (my $i=0; my @buf = caller($i); $i++) {
        $caller = $buf[1];
    }

    $mesg .= "[sent by $caller]\n";
    $mesg .= "[at $timestamp]\n";
    $mesg .= "Subject: $param{'subject'}\n" if $param{'subject'};
    $mesg .= "\n";
    $mesg .= $param{'body'};

    for my $to (@recip) {
        $self->debug("send to:".$to);
        $self->{client}->MessageSend(
            to   => $to,
            body => $mesg,
           );
    }
}

# borrow from List::MoreUtils
sub list_uniq {
    my %h;
    map { $h{$_}++ == 0 ? $_ : () } @_;
}

sub debug {
    my($self, @m) = @_;
    return unless $self->{config}{global}{debug};
    chomp @m;
    print STDERR 'DEBUG: ', @m, "\n";
}

sub DESTROY {
    my($self) = @_;
    return unless $self->{connected};
    $self->{client}->Disconnect();
}

1;
__END__

=head1 NAME

B<Scramble::XMPP> - Convenient class for send scramble message by XMPP (jabber)

=head1 SYNOPSIS

  use Scramble::XMPP;
  
  my $sc = Scramble::XMPP->new();
  $sc->send_emerg(body => "XXX is down\ntaihen! taihen!");

  $sc->send_warn( body => "XXX is okashii",
                  rcpt => ['tsuikade@jabber.jp', 'anohitonimo@jabber.jp']);

  use Encode;
  $sc->send_info( body => Encode::decode('euc-jp', 'Nihongo de OK') );

=head1 DESCRIPTION

Convenient class for send scramble message by XMPP (jabber).

=head1 METHODS

=head2 new

  my $sc = Scramble::XMPP->new(%option);

Usually it is no necessary to specify option but you can you really want.

%option is following:
(XXX) is default value.

  key         desc
  =====================================================================
  jid         sender jabber id. (scramble@jabber.jp)
  password    password of above jid. (password of scramble@jabber.jp)
  server_host jabber server name. (domain part of above jid)
  server_port jabber server port. (5222)
  tls         use TLS or not. (0)
  debug       debug mode. (0)
  config      specify config file. (/usr/klab/etc/scramble/xmpp.yaml)

priolity. upper is higher precedence.
  - option by new()
  - config YAML file.
    filename specified by new(config=>"...") or /usr/klab/etc/scramble/xmpp.yaml
  - super default value in code.

=head2 send_emerg

  $sc->send_emerg(body => "message is here.");

  $sc->send_emerg(
                  body => "message is here.",
                  rcpt => [ 'foo@jabber.jp', 'foo@jabber.jp' ],
                 );

body is required. rcpt is optional.

sender program name and timestamp is added the top of message body.

  [sent by ./test-scramblexmpp.pl]
  [at 2007-11-19 18:11:51.639074087+0900]

Current supported arguments:

=over 4

=item body => "STRING"

Set message body.
required.

=item rcpt => \@RCPT

added to default recipient list.
optional.

example:

  rcpt => ['foo@jabber.jp']

send to default recipient plus foo@jabber.jp

=item force_rcpt => \@RCPT

replace default recipient list with this.
optional

example:

  force_rcpt => ['bar@jabber.jp', 'baz@jabber.jp' ]

send to ONLY bar@jabber.jp and baz@jabber.jp.

=back

=head2 send_warn

  $sc->send_warn(body => "message is here.");

Send warning level message.

=head2 send_info

  $sc->send_info(body => "message is here.");

Send information level message.

=head1 ENVIRONMENT

=over 6

=item http_proxy

when cannot connect to XMPP server directly, you can connect via HTTP proxy.

  env http_proxy=127.0.0.1:8080 send-sc-xmpp.pl -t hirose@jabber.jp 'via proxy'

=head1 FILES

=over 4

=item F</usr/klab/etc/scramble/xmpp.yaml>

default configuration file.

=head1 AUTHOR

HIROSE, Masaaki E<lt>hirose@klab.orgE<gt>

=cut

# for Emacsen
# Local Variables:
# mode: cperl
# cperl-indent-level: 4
# indent-tabs-mode: nil
# coding: euc-jp
# End:

# vi: set ts=4 sw=4 sts=0 :
